"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const arangojs_1 = require("arangojs");
const set = require("lodash.set");
const nomatic_data_1 = require("nomatic-data");
const AdapterError_1 = require("nomatic-data/src/errors/AdapterError");
const AlreadyExistsError_1 = require("nomatic-data/src/errors/AlreadyExistsError");
const AssertionError_1 = require("nomatic-data/src/errors/AssertionError");
const NotFoundError_1 = require("nomatic-data/src/errors/NotFoundError");
const util_1 = require("util");
class ArangoDBAdapter extends nomatic_data_1.DatabaseAdapter {
    constructor(options) {
        super(options);
        const url = 'http://' + options.user + ':' + options.password + '@' + options.host + ':'
            + options.port;
        this.client = new arangojs_1.Database({
            url: url
        });
        this.name = options.name;
        this.password = options.password;
        this.user = options.user;
    }
    get name() {
        return this.client['name'];
    }
    set name(name) {
        if (!this.client) {
            return;
        }
        this.client.useDatabase(name);
    }
    set password(password) {
        this._password = password;
        if (this.user) {
            this.client.useBasicAuth(this._user, this._password);
        }
    }
    get user() {
        return this._user;
    }
    set user(user) {
        this._user = user;
        this.client['useBasicAuth'](this._user, this._password);
    }
    createCollection(collection) {
        if (this.name === '_system') {
            return Promise.reject(new Error('Cannot create a collection on ' + this.name + ' database: ' + collection));
        }
        return this.client.collection(collection).create({
            waitForSync: true
        }).then(() => {
            return true;
        }).catch((error) => {
            if (error.name !== 'ArangoError' || error.errorNum !== 1207) {
                this.handleError(error);
            }
            return false;
        });
    }
    createDatabase(database = null) {
        if (!database) {
            database = this.name;
        }
        const currentDatabaseName = this.name;
        this.name = '_system';
        return this.client.createDatabase(database, [
            {
                username: this.user
            }
        ]).then(() => {
            this.name = currentDatabaseName;
            return true;
        }).catch((error) => {
            this.name = currentDatabaseName;
            if (error.name !== 'ArangoError' || error.errorNum !== 1207) {
                this.handleError(error);
            }
            return false;
        });
    }
    decode(collection, data) {
        if (data['_id']) {
            delete data['_id'];
        }
        if (data['_key']) {
            data['id'] = data['_key'];
            delete data['_key'];
        }
        if (data['_rev']) {
            data['rev'] = data['_rev'];
            delete data['_rev'];
        }
        return data;
    }
    dropCollection(collection) {
        return __awaiter(this, void 0, void 0, function* () {
            if (collection.startsWith('_')) {
                throw new Error('Cannot delete a system collection');
            }
            try {
                yield this.client.collection(collection).drop();
                return true;
            }
            catch (error) {
                this.handleError(error);
                return false;
            }
        });
    }
    dropDatabase(database = this.name) {
        return __awaiter(this, void 0, void 0, function* () {
            if (database.startsWith('_')) {
                throw new Error('Cannot delete a system database');
            }
            const currentDatabaseName = this.name;
            this.name = '_system';
            try {
                yield this.client.dropDatabase(database);
                this.name = currentDatabaseName;
                return true;
            }
            catch (error) {
                this.name = currentDatabaseName;
                this.handleError(error);
                return false;
            }
        });
    }
    encode(collection, data) {
        if (data.hasOwnProperty('id')) {
            if (!(util_1.isNull(data.id))) {
                data._id = collection + '/' + data.id;
                data._key = data.id;
            }
            delete data.id;
        }
        if (data.hasOwnProperty('rev')) {
            if (!(util_1.isNull(data.rev))) {
                data._rev = data.rev;
            }
            delete data.rev;
        }
        return data;
    }
    ensureCollectionExists(collection) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.createCollection(collection);
        });
    }
    ensureDatabaseExists(database = this.name) {
        return this.createDatabase(database);
    }
    findAll(collection, query) {
        return __awaiter(this, void 0, void 0, function* () {
            const q = this.queryToAQL(collection, query);
            return this.client.query(this.queryToAQL(collection, query)).then((cursor) => {
                return cursor.all();
            }).then((response) => {
                const results = [];
                for (const result of response) {
                    results.push(this.decode(collection, result));
                }
                return results;
            });
        });
    }
    get(collection, id) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const response = yield this.client.collection(collection).document(id);
                return this.decode(collection, response);
            }
            catch (error) {
                this.handleError(error);
                return {};
            }
        });
    }
    getCollectionNames(database = this.name) {
        return __awaiter(this, void 0, void 0, function* () {
            const list = [];
            const currentDatabase = this.name;
            this.name = database;
            try {
                const response = yield this.client.listCollections(true);
                for (const collection of response) {
                    list.push(collection['name']);
                }
                this.name = currentDatabase;
                return list;
            }
            catch (error) {
                this.name = currentDatabase;
                this.handleError(error);
                return [];
            }
        });
    }
    getDatabaseNames() {
        const currentDatabase = this.name;
        if (this.name != '_system') {
            this.name = '_system';
        }
        return this.client.listDatabases().then((list) => {
            if (!list || !(list instanceof Array)) {
                list = [];
            }
            this.name = currentDatabase;
            return list;
        });
    }
    insert(collection, data) {
        return __awaiter(this, void 0, void 0, function* () {
            data = Object.assign({}, data);
            try {
                const response = yield this.client.collection(collection).save(this.encode(collection, data), {
                    returnNew: true,
                    waitForSync: true
                });
                data = this.decode(collection, response.new);
                return data;
            }
            catch (error) {
                this.handleError(error, collection);
                return {};
            }
        });
    }
    insertAll(collection, data) {
        const promises = [];
        for (const i in data) {
            promises.push(this.insert(collection, data[i]));
        }
        return Promise.all(promises);
    }
    queryToAQL(collection, query) {
        const lookup = {
            $and: 'AND',
            $or: 'OR',
            $eq: '==',
            $gt: '>',
            $gte: '>=',
            $in: 'ANY IN',
            $lt: '<',
            $lte: '<=',
            $ne: '!=',
            $nin: 'NOT IN'
        };
        let s = `FOR doc IN ${collection}`;
        if (query) {
            const data = query['data'];
            if (data.$where) {
                for (const logicOperator in data.$where) {
                    for (let i = 0; i < data.$where[logicOperator].length; i++) {
                        if (i === 0) {
                            s += ' FILTER ';
                        }
                        else {
                            s += ` ${lookup[logicOperator]} `;
                        }
                        for (const key in data.$where[logicOperator][i]) {
                            let aKey = key;
                            if (key === 'id') {
                                aKey = '_key';
                            }
                            else if (key === 'rev') {
                                aKey = '_rev';
                            }
                            for (const operator in data.$where[logicOperator][i][key]) {
                                let op;
                                if (operator === '$exists') {
                                    if (!data.$where[logicOperator][i][key][operator]) {
                                        op = ` == null`;
                                    }
                                    else {
                                        op = ` != null`;
                                    }
                                }
                                else {
                                    op = lookup[operator];
                                }
                                s += `doc.${aKey} ${op}`;
                                let value;
                                if (operator !== '$exists') {
                                    if (typeof data.$where[logicOperator][i][key][operator] === 'string') {
                                        value = ` "${data.$where[logicOperator][i][key][operator]}"`;
                                    }
                                    else if (data.$where[logicOperator][i][key][operator] instanceof Array) {
                                        const v = [];
                                        for (const k in data.$where[logicOperator][i][key][operator]) {
                                            if (typeof data.$where[logicOperator][i][key][operator][k] === 'string') {
                                                v.push(`"${data.$where[logicOperator][i][key][operator][k]}"`);
                                            }
                                            else {
                                                v.push(data.$where[logicOperator][i][key][operator][k]);
                                            }
                                        }
                                        value = ` [${v}]`;
                                    }
                                    else {
                                        value = ` ${data.$where[logicOperator][i][key][operator]}`;
                                    }
                                    s += value;
                                }
                            }
                        }
                    }
                }
            }
            if (data.$sort) {
                for (let i = 0; i < data.$sort.length; i++) {
                    if (i === 0) {
                        s += ' SORT ';
                    }
                    else {
                        s += ', ';
                    }
                    s += 'doc.';
                    switch (data.$sort[i][0]) {
                        case 'id':
                            s += '_key';
                            break;
                        case 'rev':
                            s += '_rev';
                            break;
                        default:
                            s += data.$sort[i][0];
                    }
                    if (data.$sort[i][1] < 0) {
                        s += ' DESC';
                    }
                }
            }
            if (data.$limit !== 0) {
                if (data.$skip) {
                    s += ` LIMIT ${data.$skip}, ${data.$limit}`;
                }
                else {
                    s += ` LIMIT ${data.$limit}`;
                }
            }
            else if (data.$skip && data.$skip !== 0) {
                s += ` LIMIT ${data.$skip}, 9999999999999`;
            }
            if (data.$fields && data.$fields.length > 0) {
                const fields = {};
                for (let i = 0; i < data.$fields.length; i++) {
                    let aKey = data.$fields[i];
                    if (data.$fields[i] === 'id') {
                        aKey = '_key';
                    }
                    else if (data.$fields[i] === 'rev') {
                        aKey = '_rev';
                    }
                    set(fields, aKey, `doc.${aKey}`);
                }
                s += ` RETURN ${util_1.inspect(fields, null, Infinity).split(`'`).join('')}`;
            }
            else {
                s += ` RETURN doc`;
            }
        }
        else {
            s += ` RETURN doc`;
        }
        return s;
    }
    remove(collection, data) {
        return __awaiter(this, void 0, void 0, function* () {
            let id;
            const options = {
                waitForSync: true,
                policy: 'error',
                returnOld: true
            };
            if (typeof data === 'string') {
                id = data;
            }
            else {
                id = data.id;
                if (data.rev) {
                    options['rev'] = data['rev'];
                }
            }
            try {
                const response = yield this.client.collection(collection).remove(id, {
                    waitForSync: true,
                    policy: 'error',
                    returnOld: true
                });
                return this.decode(collection, response.old);
            }
            catch (error) {
                this.handleError(error);
                return {};
            }
        });
    }
    replace(collection, id, data, rev = null) {
        return __awaiter(this, void 0, void 0, function* () {
            data = Object.assign({}, data);
            const options = {
                waitForSync: true,
                policy: 'error',
                returnNew: true
            };
            if (rev) {
                options['rev'] = rev;
            }
            try {
                const response = yield this.client.collection(collection).replace(id, this.encode(collection, data), options);
                return this.decode(collection, response.new);
            }
            catch (error) {
                this.handleError(error);
                return {};
            }
        });
    }
    truncateCollection(collection) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.client.collection(collection).truncate();
                return true;
            }
            catch (error) {
                this.handleError(error, collection);
                return false;
            }
        });
    }
    truncateDatabase() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.client.truncate();
                return true;
            }
            catch (error) {
                this.handleError(error);
                return false;
            }
        });
    }
    update(collection, id, data) {
        return __awaiter(this, void 0, void 0, function* () {
            data = Object.assign({}, data);
            try {
                const options = {
                    waitForSync: true,
                    keepNull: false,
                    policy: 'error',
                    returnNew: true
                };
                data = this.encode(collection, data);
                if (data['_rev']) {
                    options['rev'] = data['_rev'];
                }
                const response = yield this.client.collection(collection).update(id, data, options);
                return this.decode(collection, response.new);
            }
            catch (error) {
                if (error.name === 'ArangoError' && error['errorNum'] === 1200) {
                    throw new AssertionError_1.default('rev', data['_rev'], 'something different');
                }
                this.handleError(error);
                return {};
            }
        });
    }
    load() {
        return this.ensureDatabaseExists();
    }
    handleError(error, ...data) {
        switch (error.name) {
            case 'ArangoError':
                switch (error['errorNum']) {
                    case 1202:
                        error = new NotFoundError_1.default();
                        break;
                    case 1203:
                        const collectionName = error.message.split(': ')[1];
                        error = new AdapterError_1.default('Database `' + this.client['name'] + '` has no collection `'
                            + collectionName + '`');
                        break;
                    case 1210:
                        error = new AlreadyExistsError_1.default();
                        break;
                }
                break;
        }
        throw error;
    }
}
exports.ArangoDBAdapter = ArangoDBAdapter;
exports.default = ArangoDBAdapter;
//# sourceMappingURL=index.js.map